# combo - content management system
# Copyright (C) 2017-2023  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import json
import tarfile

from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from rest_framework import permissions
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response

from combo.apps.export_import.models import Application, ApplicationElement
from combo.data.models import Page, PageSnapshot
from combo.data.utils import import_site
from combo.utils.misc import is_portal_agent


class Index(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        if is_portal_agent():
            response = [
                {
                    'id': 'portal-agent-pages',
                    'text': _('Pages (agent portal)'),
                    'singular': _('Page (agent portal)'),
                },
            ]
        else:
            response = [
                {'id': 'pages', 'text': _('Pages'), 'singular': _('Page')},
            ]

        response[0]['urls'] = {
            'list': request.build_absolute_uri(reverse('api-export-import-pages-list')),
        }

        return Response({'data': response})


index = Index.as_view()


def get_page_bundle_entry(request, page, order):
    return {
        'id': str(page.uuid),
        'text': page.title,
        'indent': getattr(page, 'level', 0),
        'type': 'portal-agent-pages' if is_portal_agent() else 'pages',
        'order': order,
        'urls': {
            'export': request.build_absolute_uri(
                reverse('api-export-import-page-export', kwargs={'uuid': str(page.uuid)})
            ),
            'dependencies': request.build_absolute_uri(
                reverse('api-export-import-page-dependencies', kwargs={'uuid': str(page.uuid)})
            ),
            'redirect': request.build_absolute_uri(
                reverse('api-export-import-page-redirect', kwargs={'uuid': str(page.uuid)})
            ),
        },
    }


class ListPages(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        pages = Page.get_as_reordered_flat_hierarchy(Page.objects.all())
        response = [get_page_bundle_entry(request, x, i) for i, x in enumerate(pages)]
        return Response({'data': response})


list_pages = ListPages.as_view()


class ExportPage(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, uuid, *args, **kwargs):
        serialisation = Page.objects.get(uuid=uuid).get_serialized_page()
        return Response({'data': serialisation})


export_page = ExportPage.as_view()


class PageDependencies(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, uuid, *args, **kwargs):
        page = Page.objects.get(uuid=uuid)

        def dependency_dict(element):
            if isinstance(element, Group):
                return {
                    'id': element.role.slug if hasattr(element, 'role') else element.id,
                    'text': element.name,
                    'type': 'roles',
                    'urls': {},
                    # include uuid in object reference, this is not used for applification API but is useful
                    # for authentic creating its role summary page.
                    'uuid': element.role.uuid if hasattr(element, 'role') else None,
                }
            elif isinstance(element, Page):
                return get_page_bundle_entry(request, element, 0)
            return element

        dependencies = [dependency_dict(x) for x in page.get_dependencies() if x]
        return Response({'err': 0, 'data': dependencies})


page_dependencies = PageDependencies.as_view()


def page_redirect(request, uuid):
    page = get_object_or_404(Page, uuid=uuid)
    url = reverse('combo-manager-page-view', kwargs={'pk': page.pk})
    if (
        'compare' in request.GET
        and request.GET.get('application')
        and request.GET.get('version1')
        and request.GET.get('version2')
    ):
        url = '%s?version1=%s&version2=%s&application=%s' % (
            reverse('combo-manager-page-history-compare', args=[page.pk]),
            request.GET['version1'],
            request.GET['version2'],
            request.GET['application'],
        )
    return redirect(url)


class BundleCheck(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    def put(self, request, *args, **kwargs):
        tar_io = io.BytesIO(request.read())
        page_type = 'portal-agent-pages' if is_portal_agent() else 'pages'
        with tarfile.open(fileobj=tar_io) as tar:
            manifest = json.loads(tar.extractfile('manifest.json').read().decode())
            application_slug = manifest.get('slug')
            application_version = manifest.get('version_number')
            if not application_slug or not application_version:
                return Response({'data': {}})

            differences = []
            unknown_elements = []
            no_history_elements = []
            legacy_elements = []
            content_type = ContentType.objects.get_for_model(Page)
            for element in manifest.get('elements'):
                if element.get('type') != page_type:
                    continue
                try:
                    page = Page.objects.get(uuid=element['slug'])
                except Page.DoesNotExist:
                    unknown_elements.append(
                        {
                            'type': element['type'],
                            'slug': element['slug'],
                        }
                    )
                    continue
                elements_qs = ApplicationElement.objects.filter(
                    application__slug=application_slug,
                    content_type=content_type,
                    object_id=page.pk,
                )
                if not elements_qs.exists():
                    # object exists, but not linked to the application
                    legacy_elements.append(
                        {
                            'type': element['type'],
                            'slug': element['slug'],
                            # information needed here, Relation objects may not exist yet in hobo
                            'text': page.title,
                            'url': request.build_absolute_uri(
                                reverse('api-export-import-page-redirect', kwargs={'uuid': str(page.uuid)})
                            ),
                        }
                    )
                    continue
                snapshot_for_app = (
                    PageSnapshot.objects.filter(
                        page=page,
                        application_slug=application_slug,
                        application_version=application_version,
                    )
                    .order_by('timestamp')
                    .last()
                )
                if not snapshot_for_app:
                    # no snapshot for this bundle
                    no_history_elements.append(
                        {
                            'type': element['type'],
                            'slug': element['slug'],
                        }
                    )
                    continue
                last_snapshot = PageSnapshot.objects.filter(page=page).latest('timestamp')
                if snapshot_for_app.pk != last_snapshot.pk:
                    differences.append(
                        {
                            'type': element['type'],
                            'slug': element['slug'],
                            'url': '%s%s?version1=%s&version2=%s'
                            % (
                                request.build_absolute_uri('/')[:-1],
                                reverse('combo-manager-page-history-compare', args=[page.pk]),
                                snapshot_for_app.pk,
                                last_snapshot.pk,
                            ),
                        }
                    )

        return Response(
            {
                'data': {
                    'differences': differences,
                    'unknown_elements': unknown_elements,
                    'no_history_elements': no_history_elements,
                    'legacy_elements': legacy_elements,
                }
            }
        )


bundle_check = BundleCheck.as_view()


class BundleImport(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)
    install = True

    def put(self, request, *args, **kwargs):
        tar_io = io.BytesIO(request.read())
        page_type = 'portal-agent-pages' if is_portal_agent() else 'pages'
        pages = []
        with tarfile.open(fileobj=tar_io) as tar:
            manifest = json.loads(tar.extractfile('manifest.json').read().decode())
            self.application = Application.update_or_create_from_manifest(
                manifest,
                tar,
                editable=not self.install,
            )
            for element in manifest.get('elements'):
                if element.get('type') != page_type:
                    continue
                pages.append(
                    json.loads(tar.extractfile(f'{page_type}/{element["slug"]}').read().decode()).get('data')
                )
        # init cache of application elements, from manifest
        self.application_elements = set()
        # install pages
        self.do_something(pages)
        # create application elements
        self.link_objects(pages)
        # remove obsolete application elements
        self.unlink_obsolete_objects()
        return Response({'err': 0})

    def do_something(self, pages):
        if pages:
            import_site({'pages': pages})

    def link_objects(self, pages):
        for page in pages:
            page_uuid = page['fields']['uuid']
            try:
                existing_page = Page.objects.get(uuid=page_uuid)
            except Page.DoesNotExist:
                pass
            else:
                element = ApplicationElement.update_or_create_for_object(self.application, existing_page)
                self.application_elements.add(element.content_object)
                if self.install is True:
                    PageSnapshot.take(
                        existing_page,
                        request=self.request,
                        comment=_('Application (%s)') % self.application,
                        application=self.application,
                    )

    def unlink_obsolete_objects(self):
        known_elements = ApplicationElement.objects.filter(application=self.application)
        for element in known_elements:
            if element.content_object not in self.application_elements:
                element.delete()


bundle_import = BundleImport.as_view()


class BundleDeclare(BundleImport):
    install = False

    def do_something(self, pages):
        # no installation on declare
        pass


bundle_declare = BundleDeclare.as_view()


class BundleUnlink(GenericAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        if request.POST.get('application'):
            try:
                application = Application.objects.get(slug=request.POST['application'])
            except Application.DoesNotExist:
                pass
            else:
                application.delete()

        return Response({'err': 0})


bundle_unlink = BundleUnlink.as_view()
